﻿using System;
namespace LogisticWithMulti
{
  class LogisticMultiProgram
  {
    static void Main(string[] args)
    {
      Console.WriteLine("\nBegin Logistic Regression Classification with Multi-Swarm demo");

      int numFeatures = 5;
      int numRows = 10000;
      int seed = 0; 

      Console.WriteLine("\nGenerating " + numRows + " artificial data items with " + numFeatures + " features");
      double[][] allData = MakeAllData(numFeatures, numRows, seed);
      Console.WriteLine("Done");

      Console.WriteLine("\nCreating train (80%) and test (20%) matrices");
      double[][] trainData;
      double[][] testData;
      MakeTrainTest(allData, 0.80, seed, out trainData, out testData);
      Console.WriteLine("Done");

      Console.WriteLine("\nTraining data: \n");
      ShowData(trainData, 4, 2, true);

      Console.WriteLine("\nTest data: \n");
      ShowData(testData, 3, 2, true);

      Console.WriteLine("Creating Logistic Regression classifier\n");
      LogisticClassifier lc = new LogisticClassifier(numFeatures);

      int numSwarms = 4;
      int numParticles = 3;
      int maxEpochs = 100;
      Console.WriteLine("Setting numSwarms = " + numSwarms);
      Console.WriteLine("Setting numParticles = " + numParticles);
      Console.WriteLine("Setting maxEpochs = " + maxEpochs);
      Console.WriteLine("\nStarting training");
      double[] bestWeights = lc.Train(trainData, maxEpochs, numSwarms, numParticles);
      Console.WriteLine("Training complete\n");
      Console.WriteLine("Best weights found:");
      ShowVector(bestWeights, 4, true);

      double trainAcc = lc.Accuracy(trainData, bestWeights);
      Console.WriteLine("Accuracy on training data = " + trainAcc.ToString("F4"));

      double testAcc = lc.Accuracy(testData, bestWeights);
      Console.WriteLine("Accuracy on test data = " + testAcc.ToString("F4"));

      Console.WriteLine("\nEnd Logistic Classification with Multi-Swarm demo\n");
      Console.ReadLine();
    } // Main

    static double[][] MakeAllData(int numFeatures, int numRows, int seed)
    {
      Random rnd = new Random(seed);
      double[] weights = new double[numFeatures + 1]; // inc. b0
      for (int i = 0; i < weights.Length; ++i)
        weights[i] = 20.0 * rnd.NextDouble() - 10.0; // [-10.0 to +10.0]

      double[][] result = new double[numRows][]; // allocate matrix
      for (int i = 0; i < numRows; ++i)
        result[i] = new double[numFeatures + 1]; // Y in last column

      for (int i = 0; i < numRows; ++i) // for each row
      {
        double y = weights[0]; // the b0 
        for (int j = 0; j < numFeatures; ++j) // each feature / column except last
        {
          double x = 20.0 * rnd.NextDouble() - 10.0; // random X in [10.0, +10.0]
          result[i][j] = x; // store x
          double wx = x * weights[j + 1]; // weight * x 
          y += wx; // accumulate to get Y
        }

        if (y > 2.0)
          result[i][numFeatures] = 1.0; // store y in last column
        else
          result[i][numFeatures] = 0.0;
      }
      Console.WriteLine("Data generation weights:");
      ShowVector(weights, 4, false);

      return result;
    } // MakeAllData

    static void MakeTrainTest(double[][] allData, double trainPct, int seed,
      out double[][] trainData, out double[][] testData)
    {
      Random rnd = new Random(seed);
      int totRows = allData.Length;
      int numTrainRows = (int)(totRows * trainPct); // usually 0.80
      int numTestRows = totRows - numTrainRows;
      trainData = new double[numTrainRows][];
      testData = new double[numTestRows][];

      double[][] copy = new double[allData.Length][]; // ref copy of all data
      for (int i = 0; i < copy.Length; ++i)
        copy[i] = allData[i];

      for (int i = 0; i < copy.Length; ++i) // scramble order
      {
        int r = rnd.Next(i, copy.Length); // use Fisher-Yates
        double[] tmp = copy[r];
        copy[r] = copy[i];
        copy[i] = tmp;
      }
      for (int i = 0; i < numTrainRows; ++i)
        trainData[i] = copy[i];

      for (int i = 0; i < numTestRows; ++i)
        testData[i] = copy[i + numTrainRows];
    } // MakeTrainTest


    static void ShowData(double[][] data, int numRows,
      int decimals, bool indices)
    {
      for (int i = 0; i < numRows; ++i)
      {
        if (indices == true)
          Console.Write("[" + i.ToString().PadLeft(2) + "]  ");
        for (int j = 0; j < data[i].Length; ++j)
        {
          double v = data[i][j];
          if (v >= 0.0)
            Console.Write(" "); // '+'
          Console.Write(v.ToString("F" + decimals) + "    ");
        }
        Console.WriteLine("");
      }
      Console.WriteLine(". . .");
      int lastRow = data.Length - 1;
      if (indices == true)
        Console.Write("[" + lastRow.ToString().PadLeft(2) + "]  ");
      for (int j = 0; j < data[lastRow].Length; ++j)
      {
        double v = data[lastRow][j];
        if (v >= 0.0)
          Console.Write(" "); // '+'
        Console.Write(v.ToString("F" + decimals) + "    ");
      }
      Console.WriteLine("\n");
    }

    static void ShowVector(double[] vector, int decimals, bool newLine)
    {
      for (int i = 0; i < vector.Length; ++i)
        Console.Write(vector[i].ToString("F" + decimals) + " ");
      Console.WriteLine("");
      if (newLine == true)
        Console.WriteLine("");
    }

  } // Program

  public class LogisticClassifier
  {
    private int numFeatures; // number of independent variables aka features
    private double[] weights; // weights[0] = b0 = is the constant
    private Random rnd;

    public LogisticClassifier(int numFeatures)
    {
      this.numFeatures = numFeatures; // number features/predictors
      this.weights = new double[numFeatures + 1]; // [0] = b0 constant
      this.rnd = new Random(0); // pDeath, pImmigrate, sequencing
    }

    public double[] Train(double[][] trainData, int maxEpochs, int numSwarms, int numParticles)
    {
      int dim = numFeatures + 1; // plus the b0 constant
      double minX = -10.0;
      double maxX = 10.0;

      MultiSwarm ms = new MultiSwarm(numSwarms, numParticles, dim); // array of arrays of Particles
      // supply error now
      for (int i = 0; i < numSwarms; ++i) // each swarm
      {
        for (int j = 0; j < numParticles; ++j) // each particle
        {
          Particle p = ms.swarms[i].particles[j];
          p.error = Error(trainData, p.position); // add error
          p.bestError = p.error;
          Array.Copy(p.position, p.bestPosition, dim);
          if (p.error < ms.swarms[i].bestError) // swarm best?
          {
            ms.swarms[i].bestError = p.error;
            Array.Copy(p.position, ms.swarms[i].bestPosition, dim);
          }
          if (p.error < ms.bestError) // global best?
          {
            ms.bestError = p.error;
            Array.Copy(p.position, ms.bestPosition, dim);
          }
        } // j
      } // i

      int epoch = 0;
      double w = 0.729; // inertia
      double c1 = 1.49445; // particle-cogntive
      double c2 = 1.49445; // swarm-social
      double c3 = 0.3645; // multiswarm-global
      double pDeath = 1.0 / maxEpochs;      // prob of particle death
      double pImmigrate = 1.0 / maxEpochs;  // prob of particle immigration

      int[] sequence = new int[numParticles]; // process particles in random order
      for (int i = 0; i < sequence.Length; ++i)
        sequence[i] = i;

      while (epoch < maxEpochs)
      {
        ++epoch;

        if (epoch % 10 == 0 && epoch < maxEpochs) // progress
        {
          Console.Write("Epoch = " + epoch);
          Console.WriteLine("   Best error = " + ms.bestError.ToString("F4"));
        }

        for (int i = 0; i < numSwarms; ++i) // each swarm
        {
          Shuffle(sequence); // move particles in random sequence
          for (int pj = 0; pj < numParticles; ++pj) // each particle
          {
            int j = sequence[pj];
            Particle p = ms.swarms[i].particles[j]; // ref to curr particle

            // 1. update velocity
            for (int k = 0; k < dim; ++k)
            {
              double r1 = rnd.NextDouble();
              double r2 = rnd.NextDouble();
              double r3 = rnd.NextDouble();
              p.velocity[k] = (w * p.velocity[k]) +
                (c1 * r1 * (p.bestPosition[k] - p.position[k])) +
                (c2 * r2 * (ms.swarms[i].bestPosition[k] - p.position[k])) +
                (c3 * r3 * (ms.bestPosition[k] - p.position[k]));

              if (p.velocity[k] < minX)
                p.velocity[k] = minX;
              else if (p.velocity[k] > maxX)
                p.velocity[k] = maxX;
            } // k

            // 2. update position
            for (int k = 0; k < dim; ++k)
            {
              p.position[k] += p.velocity[k];
              // constrain all xi in new position
              if (p.position[k] < minX)
                p.position[k] = minX;
              else if (p.position[k] > maxX)
                p.position[k] = maxX;
            }

            // 3. update error at new position
            p.error = Error(trainData, p.position); // expensive
            // update best particle, swarm, and best global error if necessary
            if (p.error < p.bestError) // new best position for particle?
            {
              p.bestError = p.error;
              Array.Copy(p.position, p.bestPosition, dim);
            }
            if (p.error < ms.swarms[i].bestError) // new best for swarm?
            {
              ms.swarms[i].bestError = p.error;
              Array.Copy(p.position, ms.swarms[i].bestPosition, dim);
            }
            if (p.error < ms.bestError) // new global best?
            {
              ms.bestError = p.error;
              Array.Copy(p.position, ms.bestPosition, dim);
            }

            // 4. die?
            double p1 = rnd.NextDouble();
            //double pDeath = (maxProbDeath / maxEpochs) * epoch; // pDeath increases over time
            if (p1 < pDeath)
            {
              Particle q = new Particle(dim); // a replacement
              q.error = Error(trainData, q.position);
              Array.Copy(q.position, q.bestPosition, dim);
              q.bestError = q.error;

              if (q.error < ms.swarms[i].bestError) // new best swarm error by pure luck?
              {
                ms.swarms[i].bestError = q.error;
                Array.Copy(q.position, ms.swarms[i].bestPosition, dim);

                if (q.error < ms.bestError) // best global error?
                {
                  ms.bestError = q.error;
                  Array.Copy(q.position, ms.bestPosition, dim);
                }
              }

              ms.swarms[i].particles[j] = q; // die!
            } // die

            // 5. immigrate?
            double p2 = rnd.NextDouble();
            if (p2 < pImmigrate)
            {
              int ii = rnd.Next(0, numSwarms); // rnd swarm
              int jj = rnd.Next(0, numParticles); // rnd particle
              Particle q = ms.swarms[ii].particles[jj]; // q points to other

              ms.swarms[i].particles[j] = q;
              ms.swarms[ii].particles[jj] = p; // the exchange
              // p now points to other, q points to curr

              if (q.error < ms.swarms[i].bestError) // curr has new position
              {
                ms.swarms[i].bestError = q.error;
                Array.Copy(q.position, ms.swarms[i].bestPosition, dim);
              }
              if (p.error < ms.swarms[ii].bestError) // other has new position
              {
                ms.swarms[ii].bestError = p.error;
                Array.Copy(p.position, ms.swarms[ii].bestPosition, dim);
              }
            } // immigrate

          } // j
        } // i

      } // while

      return ms.bestPosition;
    }

    private void Shuffle(int[] sequence)
    {
      for (int i = 0; i < sequence.Length; ++i)
      {
        int r = rnd.Next(i, sequence.Length);
        int tmp = sequence[r];
        sequence[r] = sequence[i];
        sequence[i] = tmp;
      }
    }

    public double Error(double[][] trainData, double[] weights)
    {
      // mean squared error using supplied weights
      int yIndex = trainData[0].Length - 1; // y-value (0/1) is last column
      double sumSquaredError = 0.0;
      for (int i = 0; i < trainData.Length; ++i) // each data
      {
        double computed = ComputeOutput(trainData[i], weights);
        double desired = trainData[i][yIndex]; // ex: 0.0 or 1.0
        sumSquaredError += (computed - desired) * (computed - desired);
      }
      return sumSquaredError / trainData.Length;
    }

    public double ComputeOutput(double[] dataItem, double[] weights)
    {
      double z = 0.0;
      z += weights[0]; // start w/ the b0 constant
      for (int i = 0; i < weights.Length - 1; ++i) // data might include Y
        z += (weights[i + 1] * dataItem[i]); // skip first weight
      return 1.0 / (1.0 + Math.Exp(-z));
    }

    public int ComputeDependent(double[] dataItem, double[] weights)
    {
      double sum = ComputeOutput(dataItem, weights);

      if (sum <= 0.5) return 0;
      else return 1;
    }

    public double Accuracy(double[][] trainData, double[] weights)
    {
      int numCorrect = 0;
      int numWrong = 0;
      int yIndex = trainData[0].Length - 1;
      for (int i = 0; i < trainData.Length; ++i)
      {
        double computed = ComputeDependent(trainData[i], weights); // implicit cast
        double desired = trainData[i][yIndex]; // 0.0 or 1.0

        double closeness = 0.00000001; // 'epsilon'
        if (Math.Abs(computed - desired) < closeness)
          ++numCorrect;
        else
          ++numWrong;
      }
      return (numCorrect * 1.0) / (numWrong + numCorrect);
    }

    // ===== nested classes for MSO =====

    private class Particle
    {
      static Random rnd = new Random(0); // shared by all instances
      public double[] position;
      public double[] velocity;
      public double error;
      public double[] bestPosition;
      public double bestError;
      private double minX = -10.0;
      private double maxX = 10.0;

      public Particle(int dim)
      {
        position = new double[dim];
        velocity = new double[dim];
        bestPosition = new double[dim];
        for (int k = 0; k < dim; ++k)
        {
          position[k] = (maxX - minX) * rnd.NextDouble() + minX;
          velocity[k] = ((0.1 * maxX) - (0.1 * minX)) * rnd.NextDouble() + (0.1 * minX);
        }
        error = double.MaxValue; // temp dummy
        Array.Copy(position, bestPosition, dim); // not necessary
        bestError = error; // temp
      }
    } // Particle

    private class Swarm
    {
      // an array of particles + swarm best error
      public Particle[] particles;
      public double[] bestPosition; // of any particle
      public double bestError;

      public Swarm(int numParticles, int dim)
      {
        particles = new Particle[numParticles];
        for (int i = 0; i < numParticles; ++i)
          particles[i] = new Particle(dim); 

        bestPosition = new double[dim];
        bestError = double.MaxValue;
       } // ctor
    } // Swarm

    private class MultiSwarm
    {
      // an array of Swarms + global best error
      public Swarm[] swarms;
      public double[] bestPosition; // of any particle in any swarm
      public double bestError;

      public MultiSwarm(int numSwarms, int numParticles, int dim)
      {
        swarms = new Swarm[numSwarms];
        for (int i = 0; i < numSwarms; ++i)
          swarms[i] = new Swarm(numParticles, dim);
        bestPosition = new double[dim];
        bestError = double.MaxValue;
      } // ctor
    } // MultiSwarm
   
  } // LogisticClassifier

} // ns
